.hexagram {
    width: 120px;
    height: 150px;
    /* 6 lines x 25px per line */
    display: flex;
    flex-direction: column;
    justify-content: space-between;
    align-items: center;
    margin: 0 auto;
}

.yang-line {
    background-color: black;
    height: 20px;
    width: 100%;
}

.yin-line {
    display: flex;
    justify-content: space-between;
    align-items: center;
    width: 100%;
    height: 20px;
}

.yin-line-part {
    background-color: black;
    height: 20px;
    width: 40%;
    /* each part takes up 40% of the width */
}


.image {
    width: 50%;
    height: auto;
}